require('class')
require('memes')

local debug = true

my_waypoint = class(function(self)
  if debug then log ('my_waypoint = class(function(self) is started') end

	-- note to self: entry/exit points are represented using SimpleItemStack with health
	-- these stacks store item health as a float, not an int
	-- floats are typically perfectly accurate on ints from -16777216 to 16777216 (1 bit for sign, 24 bits for digits)
	-- self.next_transit_id = -16777216
	self.next_transit_id = 1000
	self.transit_info = {}
  if debug then log ('my_waypoint = class(function(self) is ended') end
end)

function my_waypoint:create(t)
  if debug then log ('my_waypoint:create(t) is started') end

	item_surface, item_pos, tp_surface, tp_pos, fresh, data = 
		assert(t[1] or t.item_surface),
		assert(t[2] or t.item_pos),
		assert(t[3] or t.tp_surface),
		assert(t[4] or t.tp_pos),
		t[5] or t.fresh,
		t[6] or t.data
    
  if debug then log ('my_waypoint:create(t) 100') end

	if(fresh == nil) then
		fresh = true
	end

  if debug then log ('my_waypoint:create(t) 200') end

  
	local ent = item_surface.create_entity({
		name='fre_pickup_mask',
		position={item_pos.x, item_pos.y - 0.5},
		force='neutral',
		bar=0,
	})
	ent.destructible = false
  
  if debug then log ('my_waypoint:create(t) 300') end

	local id = self.next_transit_id

  if self.next_transit_id < 0 then
    self.next_transit_id = 1 
    id = 1
  else
    self.next_transit_id = id + 1
  end

  if debug then 
    log ('my_waypoint:create(t) 350 ' .. item_surface.name)
    log ('id ' .. id) 
  end

	item_surface.create_entity{
		name='fre_item-pickup-only',
		position=item_pos,
		-- force='neutral',
		force='neutral',
		stack={name='fre_factory_transit', count=1, health=id},
	}
  
  if debug then log ('my_waypoint:create(t) 400') end

	self.transit_info[id] = {tp_surface=tp_surface, tp_pos=tp_pos, item_pos=item_pos, item_surface=item_surface, data=data, fresh=fresh}
  
  if debug then log ('my_waypoint:create(t) 500') end

	return id
end

function my_waypoint:get_tp_pos(transit_id)
	return {position=self.transit_info[transit_id].tp_pos, surface=self.transit_info[transit_id].tp_surface, data=self.transit_info[transit_id].data}
end

function my_waypoint:update_data(transit_id, data)
	self.transit_info[transit_id].data = data
end

function my_waypoint:update_item_pos(transit_id, surface, pos)
	self.transit_info[transit_id].item_pos = pos
	self.transit_info[transit_id].item_surface = surface

	local ent = surface.create_entity({
		name='fre_pickup_mask',
		position={pos.x, pos.y - 0.5},
		force='neutral',
		bar=0,
	})
	ent.destructible = false

	local ent = surface.create_entity({
		name='fre_item-pickup-only',
		position=pos,
		force='neutral',
		stack={name='fre_factory_transit', count=1, health=transit_id},
	})
	ent.destructible = false
end

function my_waypoint:update_tp_pos(transit_id, surface, pos)
	self.transit_info[transit_id].tp_surface = surface
	self.transit_info[transit_id].tp_pos = pos
end

function my_waypoint:place_world(transit_id)
	local info = self.transit_info[transit_id]

	local ent = info.item_surface.create_entity({
		name='fre_pickup_mask',
		position={info.item_pos.x, info.item_pos.y - 0.5},
		force='neutral',
		bar=0,
	})
	ent.destructible = false

	local ent = info.item_surface.create_entity({
		name='fre_item-pickup-only',
		position=info.item_pos,
		force='neutral',
		stack={name='fre_factory_transit', count=1, health=transit_id},
	})
	ent.destructible = false
end

function my_waypoint:clear_world(transit_id)
	local info = self.transit_info[transit_id]

	local mask = info.item_surface.find_entity('fre_pickup_mask', {info.item_pos.x, info.item_pos.y - 0.5})

	if(mask ~= nil and mask.valid) then
		mask.destroy()
	end
	
  local entity = info.item_surface.find_entity('fre_item-pickup-only', info.item_pos)
	if entity then 
    entity.destroy() 
  end
end

function my_waypoint:dispose(transit_id)
	local info = self.transit_info[transit_id]

	-- factories can call clear, then dispose which calls clear again
  if info then
    local item = info.item_surface.find_entity('fre_pickup_mask', {info.item_pos.x, info.item_pos.y - 0.5})
    if(item ~= nil) then
      item.destroy()
    end

    local item = info.item_surface.find_entity('fre_item-pickup-only', info.item_pos)
    if(item ~= nil) then
      item.destroy()
    end
  else
    log ('Error: no info!')
  end
	self.transit_info[transit_id] = nil
end

function my_waypoint:teleport(player)
	local item = nil
	local check = {defines.inventory.player_main, defines.inventory.player_quickbar}
	local inven = nil

	for _, inv_id in pairs(check) do
		inven = player.get_inventory(inv_id)
		item = inven.find_item_stack('fre_factory_transit')
	end

	if(item == nil) then
		error('Failed to locate factory transit (1)')
		return
	end

	local transit_id = item.health
  
	inven.remove(item.name)
	local transit = self.transit_info[transit_id]
  if not transit then 
    log('Failed to locate factory transit (2)')
    return 
  end
	local ent = transit.item_surface.create_entity({
		name='fre_item-pickup-only',
		position=transit.item_pos,
		force='neutral',
		stack={name='fre_factory_transit', count=1, health=transit_id},
	})
	ent.destructible = false

  log ('teleport the player!')
	player.teleport(transit.tp_pos, transit.tp_surface)
  log ('teleport succes?!')
  
	if(transit.fresh ~= nil) then
		transit.fresh = nil

		if(math.random(1,50) == 1) then
			local mi = math.random(1,#meme_magic)

			player.print(meme_magic[mi])
		end
	end

	return transit.data
end